/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc

  Authors: Sebastian Deorowicz and Agnieszka Debudaj-Grabysz

  Version: 1.0
  Date   : 2014-02-17
*/

#ifndef _KB_COLLECTOR_H
#define _KB_COLLECTOR_H

#include "defs.h"
#include "params.h"
#include "kmer.h"
#include "queues.h"
#include "radix.h"
#include <string>
#include <algorithm>
#include <numeric>
#include <array>
#include <vector>
#include <stdio.h>

using namespace std;

//----------------------------------------------------------------------------------
// Class collecting kmers belonging to a single bin
template <typename KMER_T, unsigned SIZE> class CKmerBinCollector {
	CMemoryMonitor *mm;
	CMemoryPool *pmm_bins;

	uint32 prefix;
	uint32 buffer_size;
	uint32 prefix_len;
	uint32 kmer_len;

	int64 mem_part_pmm_bins;

	uint32 suffix_len;
	uint32 count_len;
	KMER_T kmer_mask;
	KMER_T kmer_pck_mask;

	uint32 buffer_counter, total_counter;
	KMER_T *buffer;
	uchar *file_buffer;
	uint32 *pref_pack_cnts;
	uint32 bucket_shift;

	void Start();
	void Release();

public:
	CKmerBinCollector(CKMCParams &Params, CKMCQueues &Queues, uint32 _prefix, uint32 _buffer_size, uint32 _prefix_len, uint32 _kmer_len);
	~CKmerBinCollector();

	inline bool PutKmer(const KMER_T &x);
	void GetCompressionParams(uint32 &_count_len)
	{
		_count_len  = count_len;
	}
	bool GetBinPart(uchar *&_bin_part, uint64 &_bin_part_size, uint64 &_n_rec, bool last_one);
};

//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
//----------------------------------------------------------------------------------
// Insert a single kmer into a bin
template <typename KMER_T, unsigned SIZE> bool CKmerBinCollector<KMER_T, SIZE>::PutKmer(const KMER_T &x)
{
	buffer[buffer_counter].set(x);
	buffer[buffer_counter].mask(kmer_mask);

	return ++buffer_counter == buffer_size;
}

//----------------------------------------------------------------------------------
template <typename KMER_T, unsigned SIZE> CKmerBinCollector<KMER_T, SIZE>::CKmerBinCollector(CKMCParams &Params, CKMCQueues &Queues, uint32 _prefix, uint32 _buffer_size, uint32 _prefix_len, uint32 _kmer_len)
{
	mm			      = Queues.mm;
	pmm_bins	      = Queues.pmm_bins;
	mem_part_pmm_bins = Params.mem_part_pmm_bins;

	prefix      = _prefix;

	buffer_size	= min(_buffer_size, (uint32) 1 << 24);
	prefix_len	= _prefix_len;
	kmer_len	= _kmer_len;

	suffix_len	= kmer_len - prefix_len;
	
	kmer_mask.set_n_1(2*suffix_len);

	// Determine the best value of packing len
	uint64 storage = buffer_size * sizeof(KMER_T);
	count_len = 0;
	for(uint32 i = 0; i < min(suffix_len, (uint32) 12); ++i)
	{
		uint64 tmp_storage = 1;
		tmp_storage += buffer_size * ((suffix_len - i + 3) / 4);		// truncated kmers
		tmp_storage += ((uint64) 1 << (2*i)) * 2;						// counters

		if(tmp_storage <= storage)
		{
			storage   = tmp_storage;
			count_len = i;
		}
	}

	buffer         = NULL;
	file_buffer    = NULL;
	pref_pack_cnts = NULL;

	Start();
}

//----------------------------------------------------------------------------------
template <typename KMER_T, unsigned SIZE> CKmerBinCollector<KMER_T, SIZE>::~CKmerBinCollector()
{
	Release();
}

//----------------------------------------------------------------------------------
// Initialize data structures
template <typename KMER_T, unsigned SIZE> void CKmerBinCollector<KMER_T, SIZE>::Start()
{
	if(buffer)
		delete[] buffer;

	buffer = new KMER_T[buffer_size];

	buffer_counter = 0;
	total_counter  = 0;

	uint32 n_buckets = 1 << (count_len*2);
	pref_pack_cnts   = new uint32[n_buckets];
	bucket_shift     = (suffix_len - count_len) * 2;

	file_buffer = NULL;
}

//----------------------------------------------------------------------------------
// Release memory for buffers
template <typename KMER_T, unsigned SIZE> void CKmerBinCollector<KMER_T, SIZE>::Release()
{
	if(buffer)
	{
		delete[] buffer;
		buffer = NULL;
	}
	if(file_buffer)
	{
		delete[] file_buffer;
		file_buffer = NULL;
	}
	if(pref_pack_cnts)
	{
		delete[] pref_pack_cnts;
		pref_pack_cnts = NULL;
	}
}

//----------------------------------------------------------------------------------
// 1. Compact kmers in current buffer
// 2. Return the compacted kmers
template <typename KMER_T, unsigned SIZE> bool CKmerBinCollector<KMER_T, SIZE>::GetBinPart(uchar *&_bin_part, 
	uint64 &_bin_part_size, uint64 &_n_rec, bool last_one)
{
	uint32 n_buckets = 1 << (count_len*2);			// no. of bins
	uint32 i;

	if(!buffer_counter)
		return false;

	// Compute no. of occurrences of prefixes (to allow partial sorting of kmers and compaction)
	fill_n(pref_pack_cnts, n_buckets, 0);
	for(i = 0; i < buffer_counter; ++i)
		pref_pack_cnts[buffer[i].remove_suffix(bucket_shift)]++;

	uint32 rec_len          = (kmer_len - prefix_len + 3 - count_len) / 4;
	uint32 file_buffer_size = buffer_counter * (rec_len + KMER_T::QUALITY_SIZE) + 2 * n_buckets + 2 * sizeof(uint32);

	// Reserve memory via memory manager!
	pmm_bins->reserve(file_buffer);

	uchar *file_buffer_pos  = file_buffer;

	// Store count len in bin prt
	*file_buffer_pos++ = count_len & 0xFF;

	// Store quality size
	*file_buffer_pos++ = ((KMER_T::QUALITY_SIZE) << 4) + 2; // quality size + counters size

	// Store rec len in bin part
	*file_buffer_pos++ = (rec_len >> 8) & 0xFF;
	*file_buffer_pos++ = rec_len & 0xFF;

	// Store no. of kmers in bin part
	*file_buffer_pos++ =  buffer_counter >> 24;
	*file_buffer_pos++ = (buffer_counter >> 16) & 0xFF;
	*file_buffer_pos++ = (buffer_counter >>  8) & 0xFF;
	*file_buffer_pos++ = (buffer_counter      ) & 0xFF;

	// Store counter of packed prefixes
	for(uint32 i = 0; i < n_buckets; ++i)
	{
		*file_buffer_pos++ = (pref_pack_cnts[i] >> 8) & 0xFF;
		*file_buffer_pos++ = pref_pack_cnts[i] & 0xFF;
	}

	partial_sum(pref_pack_cnts, pref_pack_cnts+n_buckets, pref_pack_cnts);
	for(uint32 i = n_buckets-1; i > 0; --i)
		pref_pack_cnts[i] = pref_pack_cnts[i-1];
	pref_pack_cnts[0] = 0;

	// Store truncated kmers
	for(uint32 i = 0; i < buffer_counter; ++i)
	{
		uint32 x = (rec_len + KMER_T::QUALITY_SIZE) * pref_pack_cnts[buffer[i].remove_suffix(bucket_shift)]++;
		buffer[i].store(file_buffer_pos, x, rec_len);
	}

	_bin_part	   = file_buffer;
	_bin_part_size = file_buffer_size;
	_n_rec         = buffer_counter;

	total_counter += buffer_counter;
	buffer_counter = 0;

	file_buffer = NULL;

	if(last_one)
	{
		delete[] buffer;
		buffer = NULL;
	}

	return true;
}

#endif

// ***** EOF
