/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc

  Authors: Sebastian Deorowicz, Agnieszka Debudaj-Grabysz, Marek Kokot

  Version: 1.0
  Date   : 2014-02-17
*/

#ifndef _KB_SORTER_H
#define _KB_SORTER_H

#include "defs.h"
#include "params.h"
#include "kmer.h"
#include "radix.h"
#include <string>
#include <algorithm>
#include <numeric>
#include <array>
#include <vector>
#include <stdio.h>

//************************************************************************************************************
template <typename KMER_T, unsigned SIZE> class CKmerBinSorter_Impl;

//************************************************************************************************************
// CKmerBinSorter - sorting of k-mers in a bin
//************************************************************************************************************
template <typename KMER_T, unsigned SIZE> class CKmerBinSorter {
	CMemoryMonitor *mm; 
	CBinDesc *bd;
	CBinQueue *bq;
	CKmerQueue *kq;
	CMemoryPool *pmm_radix_buf, *pmm_cnts_large;
	CMemoryBins *memory_bins;

	int32 n_bins;
	int32 bin_id;
	uchar *data;
	uint64 size;
	uint64 n_rec;
	string desc;
	uint32 buffer_size;
	uint32 prefix_len;
	uint32 kmer_len;
	uint32 count_len;
	int n_omp_threads;

	uint64 n_unique, n_cutoff_min, n_cutoff_max, n_total;
	int32 cutoff_min, cutoff_max;
	int32 lut_prefix_len;
	int32 counter_max;

	KMER_T *buffer_input, *buffer_tmp, *buffer;

	void Expand();
	void Sort();

	friend class CKmerBinSorter_Impl<KMER_T, SIZE>;

public:
	CKmerBinSorter(CKMCParams &Params, CKMCQueues &Queues);
	~CKmerBinSorter();

	void ProcessBins();
};


//************************************************************************************************************
// CKmerBinSorter_Impl - implementation of k-mer type- and size-dependent functions
//************************************************************************************************************
template <typename KMER_T, unsigned SIZE> class CKmerBinSorter_Impl {
public: 
	static void Compact(CKmerBinSorter<KMER_T, SIZE> &ptr);
};

template <unsigned SIZE> class CKmerBinSorter_Impl<CKmer<SIZE>, SIZE> {
public: 
	static void Compact(CKmerBinSorter<CKmer<SIZE>, SIZE> &ptr);
};

template <unsigned SIZE> class CKmerBinSorter_Impl<CKmerQuake<SIZE>, SIZE> {
public: 
	static void Compact(CKmerBinSorter<CKmerQuake<SIZE>, SIZE> &ptr);
};


//************************************************************************************************************
// CKmerBinSorter
//************************************************************************************************************

//----------------------------------------------------------------------------------
// Assign queues and monitors
template <typename KMER_T, unsigned SIZE> CKmerBinSorter<KMER_T, SIZE>::CKmerBinSorter(CKMCParams &Params, CKMCQueues &Queues)
{
	mm            = Queues.mm;
	n_bins        = Params.n_bins;
	bd			  = Queues.bd;
	bq	          = Queues.bq;
	kq            = Queues.kq;

	pmm_radix_buf  = Queues.pmm_radix_buf;
	pmm_cnts_large = Queues.pmm_cnts_large;

	memory_bins    = Queues.memory_bins;

	cutoff_min     = Params.cutoff_min;
	cutoff_max     = Params.cutoff_max;
	counter_max    = Params.counter_max;

	lut_prefix_len = Params.lut_prefix_len;

	n_omp_threads  = Params.n_omp_threads;
}

//----------------------------------------------------------------------------------
template <typename KMER_T, unsigned SIZE> CKmerBinSorter<KMER_T, SIZE>::~CKmerBinSorter()
{

}

//----------------------------------------------------------------------------------
// Process the bins
template <typename KMER_T, unsigned SIZE> void CKmerBinSorter<KMER_T, SIZE>::ProcessBins()
{
	uint64 tmp_size;
	uint64 tmp_n_rec;
	uint32 c_disk;
	FILE *file;

	// Set no. of OMP threads for sorting
	omp_set_num_threads(n_omp_threads);

	// Effectively turns off cache when using memcpy from asmlib
	SetMemcpyCacheLimit(8);

	// Process bins
	while(!bq->completed())
	{
		// Gat bin data description to sort
		if(!bq->pop(bin_id, data, size, n_rec))
		{
			continue;
		}

		// Get bin data
		bd->read(bin_id, file, desc, c_disk, tmp_size, tmp_n_rec, buffer_size, prefix_len, kmer_len, count_len);

		// Uncompact the kmers - append truncate prefixes
		Expand();

		memory_bins->free(bin_id, CMemoryBins::mba_kmers);

		// Perfor sorting of kmers in a bin
		Sort();

		// Compact the same kmers (occurring at neighbour positions now)
		CKmerBinSorter_Impl<KMER_T, SIZE>::Compact(*this);
	}

	// Mark all the kmers are already processed
	kq->mark_completed();	
}

//----------------------------------------------------------------------------------
// Uncompact the kmers - append the truncated prefixes
template <typename KMER_T, unsigned SIZE> void CKmerBinSorter<KMER_T, SIZE>::Expand(void)
{
	uchar *data_p = data;

	uchar *raw_buffer_input, *raw_buffer_tmp;

	memory_bins->reserve(bin_id, raw_buffer_input, CMemoryBins::mba_input_array);
	memory_bins->reserve(bin_id, raw_buffer_tmp  , CMemoryBins::mba_tmp_array);

	buffer_input = (KMER_T *) raw_buffer_input;
	buffer_tmp   = (KMER_T *) raw_buffer_tmp;
	
	uint32 n_buckets = 1 << (count_len*2);
	uint32 *pref_pack_cnts;
	pmm_cnts_large->reserve(pref_pack_cnts);

	uint32 rec_len;
	uchar *part_buffer;
	
	uint64 to_read;
	uint64 bin_rec;
	for(bin_rec = 0; bin_rec < n_rec; bin_rec += to_read)
	{
		uint32 new_count_len = *data_p++;
		uint32 quality_size  = *data_p++;
		uint32 counters_size = quality_size & 0xF;
		quality_size >>= 4;
		rec_len = *data_p++;
		rec_len = (rec_len << 8) + *data_p++;

		if(new_count_len != count_len)
		{
			count_len = new_count_len;
			n_buckets = 1 << (count_len * 2);
		}

		// Read size of the counters of truncated prefixes
		to_read = *data_p++;
		to_read = (to_read << 8) + *data_p++;
		to_read = (to_read << 8) + *data_p++;
		to_read = (to_read << 8) + *data_p++;

		part_buffer = data_p;
		data_p += to_read * (rec_len + quality_size) + counters_size * n_buckets;

		uchar *file_buffer_pos = part_buffer;
		if(counters_size == 1)
		{
			for(uint32 i = 0; i < n_buckets; ++i)
			{
				pref_pack_cnts[i] = ((uint32) (*file_buffer_pos++));
			}
		}
		else if(counters_size == 2)
		{
			for(uint32 i = 0; i < n_buckets; ++i)
			{
				pref_pack_cnts[i]  = ((uint32) (*file_buffer_pos++)) << 8;
				pref_pack_cnts[i] += ((uint32) (*file_buffer_pos++));
			}
		}
		else if(counters_size == 3)
		{
			for(uint32 i = 0; i < n_buckets; ++i)
			{
				pref_pack_cnts[i]  = ((uint32) (*file_buffer_pos++)) << 16;
				pref_pack_cnts[i] += ((uint32) (*file_buffer_pos++)) << 8;
				pref_pack_cnts[i] += ((uint32) (*file_buffer_pos++));
			}
		}
		else if(counters_size == 4)
		{
			for(uint32 i = 0; i < n_buckets; ++i)
			{
				pref_pack_cnts[i]  = ((uint32) (*file_buffer_pos++)) << 24;
				pref_pack_cnts[i] += ((uint32) (*file_buffer_pos++)) << 16;
				pref_pack_cnts[i] += ((uint32) (*file_buffer_pos++)) << 8;
				pref_pack_cnts[i] += ((uint32) (*file_buffer_pos++));
			}
		}

		uint32 pref_id        = 0;
		uint32 pref_shift	  = 2 * (kmer_len - prefix_len - count_len);
		uint32 count_bits_len = 2 * count_len;

		partial_sum(pref_pack_cnts, pref_pack_cnts+n_buckets, pref_pack_cnts);

		// Expand the truncated kmers
		for(uint32 i = 0; i < to_read; ++i)
		{
			buffer_input[bin_rec + i].load(file_buffer_pos, rec_len);
			while(i >= pref_pack_cnts[pref_id])
				pref_id++;
			buffer_input[bin_rec + i].set_bits(pref_shift, count_bits_len, (uint64) pref_id);
		}
	}

	pmm_cnts_large->free(pref_pack_cnts);
}

//----------------------------------------------------------------------------------
// Sort the kmers
template <typename KMER_T, unsigned SIZE> void CKmerBinSorter<KMER_T, SIZE>::Sort()
{
	uint32 rec_len = (kmer_len - prefix_len + 3) / 4;

	// Use highly optimized parallel radix sort
	if(sizeof(KMER_T) == 8)
	{
		uint64 *_buffer_input = (uint64*) buffer_input;
		uint64 *_buffer_tmp   = (uint64*) buffer_tmp;

		RadixSort_buffer(pmm_radix_buf, _buffer_input, _buffer_tmp, n_rec, rec_len, n_omp_threads);

		if(rec_len % 2)
			buffer = (KMER_T*) _buffer_tmp;
		else
			buffer = (KMER_T*) _buffer_input;
	}
	else
	{
		uint32 *_buffer_input = (uint32*) buffer_input;
		uint32 *_buffer_tmp   = (uint32*) buffer_tmp;

		RadixSort_uint8(_buffer_input, _buffer_tmp, n_rec, sizeof(KMER_T), offsetof(KMER_T, data), SIZE*sizeof(typename KMER_T::data_t), rec_len, n_omp_threads);

		if(rec_len % 2)
			buffer = (KMER_T*) _buffer_tmp;
		else
			buffer = (KMER_T*) _buffer_input;
	}
}

//----------------------------------------------------------------------------------
// Compact the kmers - the same kmers (at neighbour positions now) are compated to a single kmer and counter
template <unsigned SIZE> void CKmerBinSorter_Impl<CKmer<SIZE>, SIZE>::Compact(CKmerBinSorter<CKmer<SIZE>, SIZE> &ptr)
{
	uint64 i;

	uint32 kmer_symbols    = ptr.kmer_len - ptr.lut_prefix_len;
	uint64 kmer_bytes      = kmer_symbols / 4;

	uint64 lut_recs        = 1 << (2*(ptr.lut_prefix_len - ptr.prefix_len));
	uint64 lut_size        = lut_recs * sizeof(uint64);
	uint64 counter_size    = min(BYTE_LOG(ptr.cutoff_max), BYTE_LOG(ptr.counter_max));

	uchar *out_buffer;
	uint64 *lut;
	uchar *raw_lut;

	ptr.memory_bins->reserve(ptr.bin_id, out_buffer, CMemoryBins::mba_suffix);
	ptr.memory_bins->reserve(ptr.bin_id, raw_lut, CMemoryBins::mba_lut);
	lut = (uint64*) raw_lut;

	uint32 out_pos = 0;
	uint32 count;
	CKmer<SIZE> *act_kmer;
	uint64 prefix = ptr.bin_id;

	ptr.n_unique     = 0;
	ptr.n_cutoff_min = 0;
	ptr.n_cutoff_max = 0;
	ptr.n_total      = 0;

	fill_n(lut, lut_recs, 0);

	if(prefix < 4)
		;
	else if((int32) prefix < ptr.n_bins/4 + ptr.n_bins/16)
		prefix /= 4;
	else
		prefix /= 16;

	if(ptr.n_rec)			// non-empty bin
	{
		act_kmer = &ptr.buffer[0];
		count = 1;

		ptr.n_total = ptr.n_rec;

		for(i = 1; i < ptr.n_rec; ++i)
		{
			if(*act_kmer == ptr.buffer[i])
				count++;
			else
			{
				if(count < (uint32) ptr.cutoff_min)
				{
					act_kmer = &ptr.buffer[i];
					ptr.n_cutoff_min++;
					ptr.n_unique++;
					count = 1;
				}
				else if(count > (uint32) ptr.cutoff_max)
				{	
					act_kmer = &ptr.buffer[i];
					ptr.n_cutoff_max++;
					ptr.n_unique++;
					count = 1;
				}
				else
				{
					if(count > (uint32) ptr.counter_max)
						count = ptr.counter_max;

					// Store compacted kmer
					for(int32 j = (int32) kmer_bytes-1; j >= 0; --j)
						out_buffer[out_pos++] = act_kmer->get_byte(j);
					for(int32 j = 0; j < (int32) counter_size; ++j)
						out_buffer[out_pos++] = (count >> (j * 8)) & 0xFF;

					lut[act_kmer->remove_suffix(2*kmer_symbols)]++;

					act_kmer = &ptr.buffer[i];
					count = 1;
					ptr.n_unique++;
				}
			}
		}

		if(count < (uint32) ptr.cutoff_min)
		{
			ptr.n_cutoff_min++;
		}
		else if(count >= (uint32) ptr.cutoff_max)
		{
			ptr.n_cutoff_max++;
		}
		else
		{
			if(count > (uint32) ptr.counter_max)
				count = ptr.counter_max;

			for(int32 j = (int32) kmer_bytes-1; j >= 0; --j)
				out_buffer[out_pos++] = act_kmer->get_byte(j);
			for(int32 j = 0; j < (int32) counter_size; ++j)
				out_buffer[out_pos++] = (count >> (j * 8)) & 0xFF;
			lut[act_kmer->remove_suffix(2*kmer_symbols)]++;
		}
		ptr.n_unique++;
	}

	// Push the sorted and compacted kmer bin to a priority queue in a form ready to be stored to HDD
	ptr.kq->push(ptr.bin_id, out_buffer, out_pos, (uchar*) lut, lut_size, ptr.n_unique, ptr.n_cutoff_min, ptr.n_cutoff_max, ptr.n_total);

	if(ptr.buffer_input)
	{
		ptr.memory_bins->free(ptr.bin_id, CMemoryBins::mba_input_array);
		ptr.memory_bins->free(ptr.bin_id, CMemoryBins::mba_tmp_array);
	}
	ptr.buffer = NULL;
}

//----------------------------------------------------------------------------------
// Compact the kmers - the same kmers (at neighbour positions now) are compated to a single kmer and counter
template <unsigned SIZE> void CKmerBinSorter_Impl<CKmerQuake<SIZE>, SIZE>::Compact(CKmerBinSorter<CKmerQuake<SIZE>, SIZE> &ptr)
{
	uint64 i;

	uint32 kmer_symbols    = ptr.kmer_len - ptr.lut_prefix_len;
	uint64 kmer_bytes      = kmer_symbols / 4;

	uint64 lut_recs        = 1 << (2*(ptr.lut_prefix_len - ptr.prefix_len));
	uint64 lut_size        = lut_recs * sizeof(uint64);

	uchar *out_buffer;
	uint64 *lut;
	uchar *raw_lut;

	ptr.memory_bins->reserve(ptr.bin_id, out_buffer, CMemoryBins::mba_suffix);
	ptr.memory_bins->reserve(ptr.bin_id, raw_lut, CMemoryBins::mba_lut);
	lut = (uint64*) raw_lut;

	uint32 out_pos = 0;
	double count;
	CKmerQuake<SIZE> *act_kmer;
	uint64 prefix = ptr.bin_id;

	ptr.n_unique     = 0;
	ptr.n_cutoff_min = 0;
	ptr.n_cutoff_max = 0;
	ptr.n_total      = 0;

	fill_n(lut, lut_recs, 0);

	if(prefix < 4)
		;
	else if((int32) prefix < ptr.n_bins/4 + ptr.n_bins/16)
		prefix /= 4;
	else
		prefix /= 16;

	act_kmer = &ptr.buffer[0];
	count = (double) act_kmer->quality;

	ptr.n_total = ptr.n_rec;

	if(ptr.n_rec)			// non-empty bin
	{
		for(i = 1; i < ptr.n_rec; ++i)
		{
			if(*act_kmer == ptr.buffer[i])
				count += ptr.buffer[i].quality;
			else
			{
				if(count < (double) ptr.cutoff_min)
				{
					act_kmer = &ptr.buffer[i];
					++ptr.n_cutoff_min;
					++ptr.n_unique;
					count = act_kmer->quality;
				}
				else if(count > (double) ptr.cutoff_max)
				{
					act_kmer = &ptr.buffer[i];
					++ptr.n_cutoff_max;
					++ptr.n_unique;
					count = act_kmer->quality;
				}
				else
				{
					if(count > (double) ptr.counter_max)
						count = (double) ptr.counter_max;

					// Store compacted kmer
					for(int32 j = (int32) kmer_bytes-1; j >= 0; --j)
						out_buffer[out_pos++] = act_kmer->get_byte(j);
					uint32 tmp;
					float f_count = (float) count;
					memcpy(&tmp, &f_count, 4);
					for(int32 j = 0; j < 4; ++j)
						out_buffer[out_pos++] = (tmp >> (j * 8)) & 0xFF;

					lut[act_kmer->remove_suffix(2*kmer_symbols)]++;

					act_kmer = &ptr.buffer[i];
					count = act_kmer->quality;
					++ptr.n_unique;
				}
			}
		}

		if(count < (double) ptr.cutoff_min)
		{
			++ptr.n_cutoff_min;
		}
		else if(count > (double) ptr.cutoff_max)
		{
			++ptr.n_cutoff_max;
		}
		else
		{
			if(count > (double) ptr.counter_max)
				count = (double) ptr.counter_max;

			for(int32 j = (int32) kmer_bytes-1; j >= 0; --j)
				out_buffer[out_pos++] = act_kmer->get_byte(j);

			uint32 tmp;
			float f_count = (float) count;
			memcpy(&tmp, &f_count, 4);
			for(int32 j = 0; j < 4; ++j)
				out_buffer[out_pos++] = (tmp >> (j * 8)) & 0xFF;

			lut[act_kmer->remove_suffix(2*kmer_symbols)]++;
		}
		++ptr.n_unique;
	}

	// Push the sorted and compacted kmer bin to a priority queue in a form ready to be stored to HDD
	ptr.kq->push(ptr.bin_id, out_buffer, out_pos, raw_lut, lut_size, ptr.n_unique, ptr.n_cutoff_min, ptr.n_cutoff_max, ptr.n_total);

	if(ptr.buffer_input)
	{
		ptr.memory_bins->free(ptr.bin_id, CMemoryBins::mba_input_array);
		ptr.memory_bins->free(ptr.bin_id, CMemoryBins::mba_tmp_array);
	}
	ptr.buffer = NULL;
}


//************************************************************************************************************
// CWKmerBinSorter - wrapper for multithreading purposes
//************************************************************************************************************
template <typename KMER_T, unsigned SIZE> class CWKmerBinSorter {
	CKmerBinSorter<KMER_T, SIZE> *kbs;

public:
	CWKmerBinSorter(CKMCParams &Params, CKMCQueues &Queues);
	~CWKmerBinSorter();

	void operator()();
};

//----------------------------------------------------------------------------------
// Constructor
template <typename KMER_T, unsigned SIZE> CWKmerBinSorter<KMER_T, SIZE>::CWKmerBinSorter(CKMCParams &Params, CKMCQueues &Queues)
{
	kbs = new CKmerBinSorter<KMER_T, SIZE>(Params, Queues);
}

//----------------------------------------------------------------------------------
// Destructor
template <typename KMER_T, unsigned SIZE> CWKmerBinSorter<KMER_T, SIZE>::~CWKmerBinSorter()
{
	delete kbs;
}

//----------------------------------------------------------------------------------
// Execution
template <typename KMER_T, unsigned SIZE> void CWKmerBinSorter<KMER_T, SIZE>::operator()()
{
	kbs->ProcessBins();
}

#endif

// ***** EOF
