/*
  This file is a part of KMC software distributed under GNU GPL 3 licence.
  The homepage of the KMC project is http://sun.aei.polsl.pl/kmc

  Authors: Sebastian Deorowicz, Agnieszka Debudaj-Grabysz, Marek Kokot

  Version: 1.0
  Date   : 2014-02-17
*/

#ifndef _QUEUES_H
#define _QUEUES_H


#include "defs.h"
#include <stdio.h>
#include <iostream>
#include <tuple>
#include <queue>
#include <list>
#include <map>
#include <string>

using namespace std;

#ifdef THREADS_NATIVE			// C++11 threads
#include <thread>
#include <mutex>
#include <condition_variable>

using namespace std::thread;
#else							// Boost threads
#include <boost/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition_variable.hpp>

using boost::thread;
#endif

#define MAX_DISKS		16

//************************************************************************************************************
class CInputFilesQueue {
	typedef string elem_t;
	typedef queue<elem_t, list<elem_t>> queue_t;

	queue_t q;
	bool is_completed;

	mutable boost::mutex mtx;								// The mutex to synchronise on

public:
	CInputFilesQueue(const vector<string> &file_names) {
		boost::unique_lock<boost::mutex> lck(mtx);

		for(vector<string>::const_iterator p = file_names.begin(); p != file_names.end(); ++p)
			q.push(*p);

		is_completed = false;
	};
	~CInputFilesQueue() {};

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty();
	}
	bool completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty() && is_completed;
	}
	void mark_completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		is_completed = true;
	}
	bool pop(string &file_name) {
		boost::lock_guard<boost::mutex> lck(mtx);

		if(q.empty())
			return false;

		file_name = q.front();
		q.pop();

		return true;
	}
};

//************************************************************************************************************
class CPartQueue {
	typedef pair<uchar *, uint64> elem_t;
	typedef queue<elem_t, list<elem_t>> queue_t;

	queue_t q;
	bool is_completed;
	int n_readers;

	mutable boost::mutex mtx;								// The mutex to synchronise on
	boost::condition_variable cv_queue_empty;

public:
	CPartQueue(int _n_readers) {
		boost::unique_lock<boost::mutex> lck(mtx);

		is_completed    = false;
		n_readers       = _n_readers;
	};
	~CPartQueue() {};

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty();
	}
	bool completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty() && !n_readers;
	}
	void mark_completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_readers--;
		if(!n_readers)
			cv_queue_empty.notify_all();
	}
	void push(uchar *part, uint64 size) {
		boost::unique_lock<boost::mutex> lck(mtx);
		bool was_empty = q.empty();
		q.push(make_pair(part, size));

		if(was_empty)
			cv_queue_empty.notify_one();
	}
	bool pop(uchar *&part, uint64 &size) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_queue_empty.wait(lck, [this]{return !this->q.empty() || !this->n_readers;}); 

		if(q.empty())
			return false;

		part = q.front().first;
		size = q.front().second;
		q.pop();

		return true;
	}
};

//************************************************************************************************************
class CWriteBufferQueue {
	typedef tuple<FILE*, uchar *, uint64> elem_t;
	typedef queue<elem_t, list<elem_t>> queue_t;

	queue_t q;
	bool is_completed;
	int n_readers;

	mutable boost::mutex mtx;								// The mutex to synchronise on
	boost::condition_variable cv_queue_empty;

public:
	CWriteBufferQueue(int _n_readers) {
		boost::unique_lock<boost::mutex> lck(mtx);

		is_completed    = false;
		n_readers       = _n_readers;
	};
	~CWriteBufferQueue() {};

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty();
	}
	bool completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty() && !n_readers;
	}
	void mark_completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_readers--;
		if(!n_readers)
			cv_queue_empty.notify_all();
	}
	void push(FILE *file, uchar *part, uint64 size) {
		boost::unique_lock<boost::mutex> lck(mtx);
		bool was_empty = q.empty();
		q.push(std::make_tuple(file, part, size));

		if(was_empty)
			cv_queue_empty.notify_one();
	}
	bool pop(FILE* &file, uchar *&part, uint64 &size) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_queue_empty.wait(lck, [this]{return !this->q.empty() || !this->n_readers;}); 

		if(q.empty())
			return false;

		file = get<0>(q.front());
		part = get<1>(q.front());
		size = get<2>(q.front());
		q.pop();

		return true;
	}
};

//************************************************************************************************************
class CBinPartQueue {
	typedef tuple<int32, uchar *, uint32, uint32> elem_t;
	typedef queue<elem_t, list<elem_t>> queue_t;
	queue_t q;

	int n_writers;
	bool is_completed;

	mutable boost::mutex mtx;						// The mutex to synchronise on
	boost::condition_variable cv_queue_empty;

public:
	CBinPartQueue(int _n_writers) {
		boost::lock_guard<boost::mutex> lck(mtx);

		n_writers       = _n_writers;
		is_completed    = false;
	}
	~CBinPartQueue() {}

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty();
	}
	bool completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty() && !n_writers;
	}
	void mark_completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_writers--;
		if(!n_writers)
			cv_queue_empty.notify_all();
	}
	void push(int32 bin_id, uchar *part, uint32 true_size, uint32 alloc_size) {
		boost::unique_lock<boost::mutex> lck(mtx);

		bool was_empty = q.empty();
		q.push(std::make_tuple(bin_id, part, true_size, alloc_size));
		if(was_empty)
			cv_queue_empty.notify_one();
	}
	bool pop(int32 &bin_id, uchar *&part, uint32 &true_size, uint32 &alloc_size) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_queue_empty.wait(lck, [this]{return !q.empty() || !n_writers;}); 

		if(q.empty())
			return false;

		bin_id     = get<0>(q.front());
		part       = get<1>(q.front());
		true_size  = get<2>(q.front());
		alloc_size = get<3>(q.front());
		q.pop();

		return true;
	}
};

//************************************************************************************************************
class CBinPackageQueue {
public:
	typedef list<tuple<uchar *, uint32, uint32>> parts_t;

private:
	typedef tuple<int32, int32, FILE*, parts_t*> elem_t;
	typedef queue<elem_t, list<elem_t>> queue_t;
	queue_t q;

	int n_writers;
	bool is_completed;

	mutable boost::mutex mtx;						// The mutex to synchronise on
	boost::condition_variable cv_queue_empty;

public:
	CBinPackageQueue(int _n_writers) {
		boost::lock_guard<boost::mutex> lck(mtx);

		n_writers       = _n_writers;
		is_completed    = false;
	}
	~CBinPackageQueue() {}

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty();
	}
	bool completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty() && !n_writers;
	}
	void mark_completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_writers--;
		if(!n_writers)
			cv_queue_empty.notify_all();
	}
	void push(int32 bin_id, int32 disk_id, FILE *bin_file, parts_t *parts, uint64 size) {
		boost::unique_lock<boost::mutex> lck(mtx);

		bool was_empty = q.empty();
		q.push(std::make_tuple(bin_id, disk_id, bin_file, parts));

		if(was_empty)
			cv_queue_empty.notify_one();
	}
	bool pop(int32 &bin_id, int32 &disk_id, FILE* &bin_file, parts_t* &parts) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_queue_empty.wait(lck, [this]{return !q.empty() || !n_writers;}); 

		if(q.empty())
			return false;
		
		bin_id   = get<0>(q.front());
		disk_id  = get<1>(q.front());
		bin_file = get<2>(q.front());
		parts    = get<3>(q.front());
		q.pop();

		return true;
	}
};

//************************************************************************************************************
class CBinDesc {
	typedef tuple<string, uint32, int64, uint64, uint32, uint32, uint32, uint32, FILE*> desc_t;
	typedef map<int32, desc_t> map_t;

	map_t m;
	int32 bin_id;

	mutable boost::mutex mtx;

public:
	CBinDesc() {
		boost::lock_guard<boost::mutex> lck(mtx);
		bin_id = -1;
	}
	~CBinDesc() {}

	void reset_reading() {
		boost::lock_guard<boost::mutex> lck(mtx);
		bin_id = -1;
	}

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return m.empty();
	}

	int32 get_next_bin()
	{
		boost::lock_guard<boost::mutex> lck(mtx);
		map_t::iterator p;
		if(bin_id == -1)
			p = m.begin();
		else
		{
			p = m.find(bin_id);
			if(p != m.end())
				++p;
		}

		if(p == m.end())
			bin_id = -1000;
		else
			bin_id = p->first;		

		return bin_id;
	}
	void insert(int32 bin_id, FILE *file, string desc, uint32 c_disk, int64 size, uint64 n_rec, uint32 buffer_size = 0, uint32 prefix_len = 0, uint32 kmer_len = 0, uint32 count_len = 0) {
		boost::lock_guard<boost::mutex> lck(mtx);

		map_t::iterator p = m.find(bin_id);
		if(p != m.end())
		{
			if(desc != "")
			{
				get<0>(m[bin_id]) = desc;
				get<1>(m[bin_id]) = c_disk;
				get<8>(m[bin_id]) = file;
			}
			get<2>(m[bin_id]) += size;
			get<3>(m[bin_id]) += n_rec;

			if(buffer_size)
			{
				get<4>(m[bin_id]) = buffer_size;
				get<5>(m[bin_id]) = prefix_len;
				get<6>(m[bin_id]) = kmer_len;
				get<7>(m[bin_id]) = count_len;
			}
		}
		else
			m[bin_id] = std::make_tuple(desc, c_disk, size, n_rec, buffer_size, prefix_len, kmer_len, count_len, file);
	}
	void read(int32 bin_id, FILE *&file, string &desc, uint32 &c_disk, uint64 &size, uint64 &n_rec, uint32 &buffer_size, uint32 &prefix_len, uint32 &kmer_len, uint32 &count_len) {
		boost::lock_guard<boost::mutex> lck(mtx);

		desc		= get<0>(m[bin_id]);
		file        = get<8>(m[bin_id]);
		c_disk      = get<1>(m[bin_id]);
		size		= (uint64) get<2>(m[bin_id]);
		n_rec		= get<3>(m[bin_id]);
		buffer_size = get<4>(m[bin_id]);
		prefix_len  = get<5>(m[bin_id]);
		kmer_len    = get<6>(m[bin_id]);
		count_len   = get<7>(m[bin_id]);
	}
	void read(int32 bin_id, FILE *&file, string &desc, uint32 &c_disk, uint64 &size, uint64 &n_rec) {
		boost::lock_guard<boost::mutex> lck(mtx);

		desc		= get<0>(m[bin_id]);
		file        = get<8>(m[bin_id]);
		c_disk		= get<1>(m[bin_id]);
		size		= (uint64) get<2>(m[bin_id]);
		n_rec		= get<3>(m[bin_id]);
	}
};

//************************************************************************************************************
class CBinQueue {
	typedef tuple<int32, uchar *, uint64, uint64> elem_t;
	typedef queue<elem_t, list<elem_t>> queue_t;
	queue_t q;

	int n_writers;

	mutable boost::mutex mtx;								// The mutex to synchronise on
	boost::condition_variable cv_queue_empty;

public:
	CBinQueue(int _n_writers) {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_writers = _n_writers;
	}
	~CBinQueue() {}

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty();
	}
	bool completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return q.empty() && !n_writers;
	}
	void mark_completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_writers--;
		if(n_writers == 0)
			cv_queue_empty.notify_all();
	}
	void push(int32 bin_id, uchar *part, uint64 size, uint64 n_rec) {
		boost::lock_guard<boost::mutex> lck(mtx);
		bool was_empty = q.empty();
		q.push(std::make_tuple(bin_id, part, size, n_rec));
		if(was_empty)
			cv_queue_empty.notify_one();
	}
	bool pop(int32 &bin_id, uchar *&part, uint64 &size, uint64 &n_rec) {
		boost::unique_lock<boost::mutex> lck(mtx);

		cv_queue_empty.wait(lck, [this]{return !q.empty() || !n_writers;}); 

		if(q.empty())
			return false;

		bin_id = get<0>(q.front());
		part   = get<1>(q.front());
		size   = get<2>(q.front());
		n_rec  = get<3>(q.front());
		q.pop();

		return true;
	}
};

//************************************************************************************************************
class CKmerQueue {
	typedef tuple<int32, uchar*, uint64, uchar*, uint64, uint64, uint64, uint64, uint64> data_t;
	typedef list<data_t> list_t;

	int n_writers;

	mutable boost::mutex mtx;								// The mutex to synchronise on
	boost::condition_variable cv_queue_empty;

	list_t l;
	int32 n_bins;
	int32 cur_id;

public:
	CKmerQueue(int32 _n_bins, int _n_writers) {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_bins = _n_bins;
		cur_id = 0;

		n_writers = _n_writers;
	}
	~CKmerQueue() {
	}

	bool empty() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return cur_id >= n_bins;
	}
	bool ready() {
		boost::lock_guard<boost::mutex> lck(mtx);
		if(l.empty())
			return false;
		return get<0>(l.front()) == cur_id;
	}
	bool completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		return l.empty() && !n_writers;
	}
	void mark_completed() {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_writers--;
		if(!n_writers)
			cv_queue_empty.notify_all();
	}
	void push(int32 bin_id, uchar *data, uint64 data_size, uchar *lut, uint64 lut_size, uint64 n_unique, uint64 n_cutoff_min, uint64 n_cutoff_max, uint64 n_total) {
		boost::lock_guard<boost::mutex> lck(mtx);

		list_t::iterator p;
		for(p = l.begin(); p != l.end(); ++p)
			if(bin_id < get<0>(*p))
			{
				l.insert(p, std::make_tuple(bin_id, data, data_size, lut, lut_size, n_unique, n_cutoff_min, n_cutoff_max, n_total));
				break;
			}
		if(p == l.end())
		{
			l.insert(p, std::make_tuple(bin_id, data, data_size, lut, lut_size, n_unique, n_cutoff_min, n_cutoff_max, n_total));
		}

		cv_queue_empty.notify_all();
	}
	bool pop(int32 &bin_id, uchar *&data, uint64 &data_size, uchar *&lut, uint64 &lut_size, uint64 &n_unique, uint64 &n_cutoff_min, uint64 &n_cutoff_max, uint64 &n_total) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_queue_empty.wait(lck, [this]{return (!l.empty() && get<0>(l.front()) == cur_id) || !n_writers;}); 
		if(l.empty())
			return false;

		bin_id       = get<0>(l.front());
		data         = get<1>(l.front());
		data_size    = get<2>(l.front());
		lut          = get<3>(l.front());
		lut_size     = get<4>(l.front());
		n_unique     = get<5>(l.front());
		n_cutoff_min = get<6>(l.front());
		n_cutoff_max = get<7>(l.front());
		n_total      = get<8>(l.front());
				
		l.pop_front();

		if(cur_id < 4)
			cur_id++;
		else if(cur_id < n_bins/4 + n_bins/16)
			cur_id += 4;
		else
			cur_id += 16;

		if(cur_id >= n_bins)
			cv_queue_empty.notify_all();

		return true;
	}
};

//************************************************************************************************************
class CMemoryMonitor {
	uint64 max_memory;
	uint64 memory_in_use;

	mutable boost::mutex mtx;								// The mutex to synchronise on
	boost::condition_variable cv_memory_full;				// The condition to wait for

public:
	CMemoryMonitor(uint64 _max_memory) {
		boost::lock_guard<boost::mutex> lck(mtx);
		max_memory    = _max_memory;
		memory_in_use = 0;
	}
	~CMemoryMonitor() {
	}

	void increase(uint64 n) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_memory_full.wait(lck, [this, n]{return memory_in_use + n <= max_memory;});
		memory_in_use += n;
	}
	void force_increase(uint64 n) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_memory_full.wait(lck, [this, n]{return memory_in_use + n <= max_memory || memory_in_use == 0;});
		memory_in_use += n;
	}
	void decrease(uint64 n) {
		boost::lock_guard<boost::mutex> lck(mtx);
		memory_in_use -= n;
		cv_memory_full.notify_all();
	}
	void info(uint64 &_max_memory, uint64 &_memory_in_use)
	{
		boost::lock_guard<boost::mutex> lck(mtx);
		_max_memory    = max_memory;
		_memory_in_use = memory_in_use;
	}
};

//************************************************************************************************************
class CDiskMonitor {
	uint32 n_disks;
	mutable boost::mutex v_mtx[MAX_DISKS];
	boost::condition_variable v_cv[MAX_DISKS];
	bool v_disk[MAX_DISKS];

public:
	CDiskMonitor(uint32 _n_disks) {
		n_disks = NORM(_n_disks, 1, MAX_DISKS);
		for(uint32 i = 0; i < n_disks; ++i)
		{
			boost::lock_guard<boost::mutex> lck(v_mtx[i]);
			v_disk[i] = true;
		}
	}
	~CDiskMonitor() {
	}

	void block(uint32 n) {
		boost::unique_lock<boost::mutex> lck(v_mtx[n]);
		v_cv[n].wait(lck, [this, n]{return v_disk[n];});
		v_disk[n] = false;
	}
	void unblock(uint32 n) {
		boost::lock_guard<boost::mutex> lck(v_mtx[n]);
		v_disk[n] = true;
		v_cv[n].notify_all();
	}
};

//************************************************************************************************************
class CBinOrdering {
	int32 n_bins;
	int32 cur_bin;

	mutable boost::mutex mtx;							// The mutex to synchronise on
	boost::condition_variable cv_bin;					// The condition to wait for

	int bin_no_skip(int cur, int tot) {
		if(cur < 4)
			return 1;
		else if(cur < tot/4 + tot/16)
			return 4;
		else
			return 16;
	}

public:
	CBinOrdering(int32 _n_bins) {
		boost::lock_guard<boost::mutex> lck(mtx);
		n_bins  = _n_bins;
		cur_bin = -1;
	}
	~CBinOrdering() {
	}
	void block(int32 n) {
		boost::unique_lock<boost::mutex> lck(mtx);
		cv_bin.wait(lck, [this, n]{return this->cur_bin + bin_no_skip(cur_bin, n_bins) >= n;});
	}
	void unblock() {
		boost::lock_guard<boost::mutex> lck(mtx);
		cur_bin += bin_no_skip(cur_bin, n_bins);
		cv_bin.notify_all();
	}
};

//************************************************************************************************************
class CMemoryPool {
	int64 total_size;
	int64 part_size;
	int64 n_parts_total;
	int64 n_parts_free;

	uchar *buffer, *raw_buffer;
	uint32 *stack;

	mutable boost::mutex mtx;							// The mutex to synchronise on
	boost::condition_variable cv;						// The condition to wait for

public:
	CMemoryPool(int64 _total_size, int64 _part_size) {
		raw_buffer = NULL;
		buffer = NULL;
		stack  = NULL;
		prepare(_total_size, _part_size);
	}
	~CMemoryPool() {
		release();
	}

	void prepare(int64 _total_size, int64 _part_size) {
		release();

		n_parts_total = _total_size / _part_size;
		part_size     = (_part_size + 15) / 16 * 16;			// to allow mapping pointer to int*
		n_parts_free  = n_parts_total;

		total_size = n_parts_total * part_size;

		raw_buffer = new uchar[total_size+64];
		buffer     = raw_buffer;
		while(((uint64) buffer) % 64)
			buffer++;

		stack = new uint32[n_parts_total];
		for(uint32 i = 0; i < n_parts_total; ++i)
			stack[i] = i;
	}

	void release(void) {
		if(raw_buffer)
			delete[] raw_buffer;
		raw_buffer = NULL;
		buffer     = NULL;

		if(stack)
			delete[] stack;
		stack = NULL;
	}

	// Allocate memory buffer - uchar*
	void reserve(uchar* &part)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		cv.wait(lck, [this]{return n_parts_free > 0;});

		part = buffer + stack[--n_parts_free]*part_size;
	}
	// Allocate memory buffer - char*
	void reserve(char* &part)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		cv.wait(lck, [this]{return n_parts_free > 0;});

		part = (char*) (buffer + stack[--n_parts_free]*part_size);
	}
	// Allocate memory buffer - uint32*
	void reserve(uint32* &part)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		cv.wait(lck, [this]{return n_parts_free > 0;});

		part = (uint32*) (buffer + stack[--n_parts_free]*part_size);
	}
	// Allocate memory buffer - uint64*
	void reserve(uint64* &part)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		cv.wait(lck, [this]{return n_parts_free > 0;});

		part = (uint64*) (buffer + stack[--n_parts_free]*part_size);
	}
	// Allocate memory buffer - double*
	void reserve(double* &part)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		cv.wait(lck, [this]{return n_parts_free > 0;});

		part = (double*) (buffer + stack[--n_parts_free]*part_size);
	}

	// Deallocate memory buffer - uchar*
	void free(uchar* part)
	{
		boost::lock_guard<boost::mutex> lck(mtx);

		stack[n_parts_free++] = (uint32) ((part - buffer) / part_size);
		cv.notify_one();		
	}
	// Deallocate memory buffer - char*
	void free(char* part)
	{
		boost::lock_guard<boost::mutex> lck(mtx);

		stack[n_parts_free++] = (uint32) (((uchar*) part - buffer) / part_size);
		cv.notify_one();		
	}
	// Deallocate memory buffer - uint32*
	void free(uint32* part)
	{
		boost::lock_guard<boost::mutex> lck(mtx);

		stack[n_parts_free++] = (uint32) ((((uchar *) part) - buffer) / part_size);
		cv.notify_one();		
	}
	// Deallocate memory buffer - uint64*
	void free(uint64* part)
	{
		boost::lock_guard<boost::mutex> lck(mtx);

		stack[n_parts_free++] = (uint32) ((((uchar *) part) - buffer) / part_size);
		cv.notify_one();		
	}
	// Deallocate memory buffer - double*
	void free(double* part)
	{
		boost::lock_guard<boost::mutex> lck(mtx);

		stack[n_parts_free++] = (uint32) ((((uchar *) part) - buffer) / part_size);
		cv.notify_one();		
	}
};

//************************************************************************************************************
class CMemoryBins {
	int64 total_size;
	int64 free_size;

	uint32 n_bins;

	typedef tuple<uchar*, uchar*, uchar*, uchar*, uchar*, uchar*, int64> bin_ptrs_t;

public:
	typedef enum{mba_kmers, mba_input_array, mba_tmp_array, mba_lut, mba_suffix} mba_t;

private:
	uchar *buffer, *raw_buffer;
	bin_ptrs_t *bin_ptrs;

	list<pair<uint64, uint64>> list_reserved;
	list<pair<uint32, uint64>> list_insert_order;

	mutable boost::mutex mtx;							// The mutex to synchronise on
	boost::condition_variable cv;						// The condition to wait for

public:
	CMemoryBins(int64 _total_size, uint32 _n_bins) {
		raw_buffer = NULL;
		buffer     = NULL;
		bin_ptrs   = NULL;
		prepare(_total_size, _n_bins);
	}
	~CMemoryBins() {
		release();
	}

	int64 round_up_to_alignment(int64 x)
	{
		return (x + ALIGNMENT-1) / ALIGNMENT * ALIGNMENT;
	}

	void prepare(int64 _total_size, uint32 _n_bins) {
		release();

		n_bins = _n_bins;
		bin_ptrs = new bin_ptrs_t[n_bins];

		total_size = round_up_to_alignment(_total_size - n_bins * sizeof(bin_ptrs_t));
		free_size  = total_size;

		raw_buffer = (uchar*) malloc(total_size+ALIGNMENT);
		buffer     = raw_buffer;
		while(((uint64) buffer) % ALIGNMENT)
			buffer++;
			
		list_reserved.clear();
		list_insert_order.clear();
		list_reserved.push_back(make_pair(total_size, 0));		// guard
	}

	void release(void) {
		if(raw_buffer)
			::free(raw_buffer);
		raw_buffer = NULL;
		buffer     = NULL;

		if(bin_ptrs)
			delete[] bin_ptrs;
		bin_ptrs = NULL;
	}

	// Prepare memory buffer for bin of given id
	void init(uint32 bin_id, uint32 sorting_phases, int64 file_size, int64 kmers_size, int64 out_buffer_size, int64 lut_size)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		int64 part1_size;
		int64 part2_size;

		if(sorting_phases % 2 == 0)
		{
			part1_size = kmers_size;
			part2_size = max(max(kmers_size, file_size), lut_size + out_buffer_size);
		}
		else
		{
			part1_size = max(kmers_size, file_size);
			part2_size = max(kmers_size, lut_size + out_buffer_size);
		}

//		int64 req_size = max(kmers_size * 2, kmers_size + out_buffer_size + lut_size);
		int64 req_size = part1_size + part2_size;
		uint64 found_pos;
		uint64 last_found_pos;

		// Look for space to insert
		cv.wait(lck, [&]() -> bool{
			found_pos = total_size;
			if(!list_insert_order.empty())
			{
				last_found_pos = list_insert_order.back().second;
				for(auto p = list_reserved.begin(); p != list_reserved.end(); ++p)
					if(p->first == last_found_pos)
					{
						uint64 last_end_pos = p->first + p->second;
						++p;
						if(last_end_pos + req_size <= p->first)
						{
							found_pos = last_end_pos;
							return true;
						}
						else
							break;
					}
			}

			uint64 prev_end_pos = 0;
				
			for(auto p = list_reserved.begin(); p != list_reserved.end(); ++p)
			{
				if(prev_end_pos + req_size <= p->first)
				{
					found_pos = prev_end_pos;
					return true;
				}
				prev_end_pos = p->first + p->second;
			}

			// Reallocate memory for buffer if necessary
			if(list_insert_order.empty() && req_size > (int64) list_reserved.back().first)
			{
				::free(raw_buffer);
				total_size = round_up_to_alignment(req_size);
				free_size  = total_size;

				raw_buffer = (uchar*) malloc(total_size+ALIGNMENT);
				buffer     = raw_buffer;
				while(((uint64) buffer) % ALIGNMENT)
					buffer++;

				list_reserved.back().first = total_size;
				found_pos = 0;
				return true;
			}

			return false;
		});

		// Reserve found free space
		list_insert_order.push_back(make_pair(bin_id, found_pos));
		for(auto p = list_reserved.begin(); p != list_reserved.end(); ++p)
			if(found_pos < p->first)
			{
				list_reserved.insert(p, make_pair(found_pos, req_size));
				break;
			}

		uchar *base_ptr = get<0>(bin_ptrs[bin_id]) = buffer + found_pos;

		if(sorting_phases % 2 == 0)				// the result of sorting is in the same place as input
		{
			get<2>(bin_ptrs[bin_id]) = base_ptr;												// array with kmers
			get<3>(bin_ptrs[bin_id]) = base_ptr + part1_size;									// temporary array for sorting kmers
			get<1>(bin_ptrs[bin_id]) = get<3>(bin_ptrs[bin_id]);								// compressed bin data
		}
		else
		{
			get<2>(bin_ptrs[bin_id]) = base_ptr + part1_size;									// array with kmers
			get<3>(bin_ptrs[bin_id]) = base_ptr;												// temporary array for sorting kmers
			get<1>(bin_ptrs[bin_id]) = base_ptr;												// compressed bin data
		}
		get<4>(bin_ptrs[bin_id]) = base_ptr + part1_size;										// data for lut
		get<5>(bin_ptrs[bin_id]) = get<4>(bin_ptrs[bin_id]) + lut_size;							// data for suffixes

		free_size -= req_size;
		get<6>(bin_ptrs[bin_id]) = req_size;
	}

	void reserve(uint32 bin_id, uchar* &part, mba_t t)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		if(t == mba_kmers)
			part = get<1>(bin_ptrs[bin_id]);
		else if(t == mba_input_array)
			part = get<2>(bin_ptrs[bin_id]);
		else if(t == mba_tmp_array)
			part = get<3>(bin_ptrs[bin_id]);
		else if(t == mba_lut)
			part = get<4>(bin_ptrs[bin_id]);
		else if(t == mba_suffix)
			part = get<5>(bin_ptrs[bin_id]);
	}

	// Deallocate memory buffer - uchar*
	void free(uint32 bin_id, mba_t t)
	{
		boost::unique_lock<boost::mutex> lck(mtx);
		if(t == mba_kmers)
			get<1>(bin_ptrs[bin_id]) = NULL;
		else if(t == mba_input_array)
			get<2>(bin_ptrs[bin_id]) = NULL;
		else if(t == mba_tmp_array)
			get<3>(bin_ptrs[bin_id]) = NULL;
		else if(t == mba_lut)
			get<4>(bin_ptrs[bin_id]) = NULL;
		else if(t == mba_suffix)
			get<5>(bin_ptrs[bin_id]) = NULL;

		if(!get<1>(bin_ptrs[bin_id]) && !get<2>(bin_ptrs[bin_id]) && !get<3>(bin_ptrs[bin_id]) && !get<4>(bin_ptrs[bin_id]) && !get<5>(bin_ptrs[bin_id]))
		{
			for(auto p = list_reserved.begin(); p != list_reserved.end() && p->second != 0; ++p)
			{
				if((int64) p->first == get<0>(bin_ptrs[bin_id]) - buffer)
				{
					list_reserved.erase(p);
					break;
				}
			}
			for(auto p = list_insert_order.begin(); p != list_insert_order.end(); ++p)
				if(p->first == bin_id)
				{
					list_insert_order.erase(p);
					break;
				}

			get<0>(bin_ptrs[bin_id]) = NULL;
			free_size += get<6>(bin_ptrs[bin_id]);
			cv.notify_all();		
		}
	}
};

#endif

// ***** EOF
